#! perl

# read in a haplotype or genotype file
# BlockID	NumberOfHaplotypes	SNPs
# SNP: 0-major, 1-minor, n-incomplete

$NumberOfMissingSNPs = $ARGV[0];
$filename = $ARGV[1];

#for($NumberOfMissingSNPs = 1; $NumberOfMissingSNPs <11; $NumberOfMissingSNPs++){
$NumOfIteration = 10;

$TotalUsedSNPs = 0;
$blockcount = 0;
$Feasibleblock = 0;
$ResultFile = "ILP_".$NumberOfMissingSNPs.".txt";

#print "$NumberOfMissingSNPs!!! \n";

unless (open (Result, ">$ResultFile"))
{
	print "Can not open the file $ResultFile !\n";
	exit;
}


$line=(<>);
while ($line)
{	
	#Begin of a new haplotype block
	$blockcount++;
	
	@a = split /\s+/, $line;
	
	$BlockID = $a[0];
	$n_snp = length $a[2];
	$MinNumOfSNPs = $n_snp;
	
	$i=0;
	$j=0;
	
	for ($i=0; $i<$n_snp; $i++)
	{
		$hap[$j][$i]=substr($a[2], $i, 1);      # haplotype data
	}
	
	$line=(<>);
	while($line){
		@NextArray = split /\s+/, $line;
		
		if($NextArray[0] eq $a[0]){
			#find another pattern in the same block, add into the hap matrix
			$j++;
			
			for ($i=0; $i<$n_snp; $i++)
			{
				$hap[$j][$i]=substr($NextArray[2], $i, 1);	# haplotype data
			}
			
			#Read the next line
			$line=(<>);
		
		}else{
		
			#break the while loop
			last;
		}

	}#end of the inner while loop
	
	$n_hap = $j+1;

	if($n_hap > 1)
	{
		#this is a haplotype block with more than one haplotypes
		$Feasibleblock++;

		#run the iterative LP-relaxation algorithm for several iterations
		for($iteration = 0; $iteration < $NumOfIteration; $iteration++)
		{
			# converting hap data into a lp file
			# run lp_solve.exe and return the results
			# round the results into intergers
			# check satisfactions
			# redo
			#
			
			#Quite strange, can not set srand, otherwise rand will be always the same.
			#srand(time|$$);  # random seed
			
			# initialize the satisfaction of separating two haplotypes
			
			$count_sat=0;
			for ($i=0; $i<$n_hap; $i++)
			{
			    for ($j=$i+1; $j<$n_hap; $j++)
			    {
			    	# this program can handle identical haplotype patterns
			    	# because sat[i][j] are set to 1 by default.
			        $sat[$i][$j]=1;
			        # number of SNPs required
			        $edges[$i][$j]=$NumberOfMissingSNPs + 1;
			        
			        for ($k=0; $k<$n_snp; $k++)
			        {
			            if ((($hap[$i][$k] eq "0") && ($hap[$j][$k] eq "1")) || 
			            	(($hap[$i][$k] eq "1") && ($hap[$j][$k] eq "0")))
			            {
			            	#This is able to handle two haplotypes with same patterns
			            	#Since they will not be set to 0 under this condition
			                $sat[$i][$j]=0;
			            } 
			        }
			    }
			}
			
			# final results in array z[]
			
			for ($i=0; $i<$n_snp; $i++)
			{
			    $z[$i]=0;
			}
			
			$n=1;
			
			#Run until a solution is found
			while(1)
			{
			
				# output to a lp file
				
				$fname = $filename."-lp-$n";
				
				generate_file($fname);
				
				$outputfile = $fname."-res";
				`lp_solve.exe $fname > $outputfile`;    # run the linear programming
				
				@y = parsing ($outputfile);
				
				if($y[0] eq "infeasible"){
					print "The $blockcount-th block has no solutions.\n";
					last;
				}
				# randomized rounding
				
				for ($i=0; $i<$n_snp; $i++)
				{
					# if $x[$i] is already set to 1, no need to reset it.
				#	if($z[$i]==1){
				#		$x[$i]=1;
				#	}else{
				    		$x[$i]=0;
				#    	}
				    if ($y[$i]>0)
				    {
				        $prob = rand(1000000);
				        #print "$prob\n";
				        
				        if ( $prob < $y[$i]*1000000)
				        {
				        	#temp selection is x[]
				            $x[$i]=1;
				            #Final selection is z[]
				            #$z[$i]=1;
				            #print "\n   y[", $i+1, "]=", $y[$i];
				            #print ",\t x[", $i+1, "]=1";
				        }
				    }
				}
				
				# update satisfaction
				
				for ($k=0; $k<$n_snp; $k++)
				{
					#print "$x[$k], $z[$k]\n";
				    if (($x[$k] == 1) && ($z[$k]==0))
				    {
				        for ($i=0; $i<$n_hap; $i++)
				        {
				            for ($j=$i+1; $j<$n_hap; $j++)
				            {
				                if ((($hap[$i][$k] eq "0") && ($hap[$j][$k] eq "1")) || 
				                	(($hap[$i][$k] eq "1") && ($hap[$j][$k] eq "0")))   
				         	    {
				         	 		#Final selection is z[]
				            		$z[$k]=1;

				                	$edges[$i][$j]=$edges[$i][$j]-1;
				                	
				                	if($edges[$i][$j]==0){
				                    	$sat[$i][$j]=1; 
				                    }
				                }
				            }
				        }
				    }
				}
				
				# check satisfaction
				$count=0;
				for ($i=0; $i<$n_hap; $i++)
				{
				    for ($j=$i+1; $j<$n_hap; $j++)
				    {
				        if ($sat[$i][$j]==0) 
				        {
				            $count++;
				            #print "\n   NOT satisfied pairs: (", $i+1, ",", $j+1, ")";    
				        }
				    }
				}
				
				if ($count==0)
				{
					$NumberOfSNPs = 0;
				    print "\n\nDONE! ";
				    print "Selected SNPs: ";
				    for ($k=0; $k<$n_snp; $k++)
				    {
				        if ($z[$k]==1)
				        {
				        	$NumberOfSNPs = $NumberOfSNPs +1;
				            print $k+1, " ";
				        }
				    }
				    print "\nThe $blockcount-th block uses SNPs: $NumberOfSNPs \n";
				    
				    #break the while loop
				    #go on to check the next haplotype block
				    last;
				}
				else
				{
				    $n++;
				}
			
			}  # end of while: a solution is not found
			
			if($y[0] eq "infeasible")
			{
				#This block has no solutions
				#break the for loop and minus the default number of SNPs
				$Feasibleblock--;
				#$TotalUsedSNPs = $TotalUsedSNPs - $MinNumOfSNPs;
				last;
			}
			
			if($NumberOfSNPs < $MinNumOfSNPs)
			{
				$MinNumOfSNPs = $NumberOfSNPs;
			}	
		}#end of each iteration
		
		if(!($y[0] eq "infeasible"))
		{
			$TotalUsedSNPs = $TotalUsedSNPs + $MinNumOfSNPs;
			print Result "$BlockID: $MinNumOfSNPs\n"
		}
	}#end of if: there are more than one haplotypes in this block
	else
	{
		print "$BlockID has only one haplotype!\n"
	}

}#end of the outer while loop: there are other haplotype blocks


###############################################################
################Execution Finish###############################

print "\n********************************\n";
print "Total blocks: $blockcount \n";
print "Total feasible blocks: $Feasibleblock \n";
print "Total SNPs: $TotalUsedSNPs \n";

print Result "\n********************************\n";
print Result "Total blocks: $blockcount \n";
print Result "Total feasible blocks: $Feasibleblock \n";
print Result "Total SNPs: $TotalUsedSNPs \n";
    
close Result;

#}
exit;

###############################################################
###############################################################


###############################################################
###################### Sub Functions ##########################
###############################################################

# converting a haplotype file in $hap[][] into a lp format file in fname
# write to a file that can be read by the lp_solve.exe
# First line: objective function (default is maximization)
# Example:  3x1 + 2x2;
# Second line and after: constraints
# Format: 3x1 + 2x2 - 4x3 >= 2;
# Format: int x1;
 
sub generate_file {

    my($fname) = @_;
    my $i;
    my $j;
    my $k;
    my $first_var;

    #print "\n\n", "Write to lp_solve file: $fname";
    unless (open (HAPFILE, ">$fname"))
    {
        print "Can not open the file $fname !\n";
        exit;
    }

    # print objective

    print HAPFILE "-x_1";
    for ($k=1; $k<$n_snp; $k++)
    {
        print HAPFILE " + -x_", $k+1;
    }
    print HAPFILE ";\n";

    # print variable constraints

    for ($k=0; $k<$n_snp; $k++)
    {
        print HAPFILE "x_", $k+1, " > 0\;\n";
        print HAPFILE "x_", $k+1, " <= 1\;\n";
    }

    # print constraints

    for ($i=0; $i<$n_hap; $i++)
    {
        for ($j=$i+1; $j<$n_hap; $j++)
        {
        	#each pair of haplotype patterns
	        if ($sat[$i][$j]==0)  
	        {                   # for every pair of haplotypes $i and $j that are NOT satisfied
	            $first_var = 1;
	            for ($k=0; $k<$n_snp; $k++)
	            {
	                if ((($hap[$i][$k] eq "0") && ($hap[$j][$k] eq "1")) || 
	                	(($hap[$i][$k] eq "1") && ($hap[$j][$k] eq "0")))
	                {  
	                	#they are different at the k-th SNP locus
	                	#this SNP must be add into the variables
	                	#this is able to handle two haplotypes with same patterns
	                    if ($first_var==1)
	                    {
	                    	#print the first variable
	                        print HAPFILE "   x_", $k+1;
	                        $first_var=0;
	                    }
	                    elsif (1)
	                    {
	                        print HAPFILE " + x_", $k+1;
	                    }
	                }  
	            }
	            if ($first_var==0) 
	            {
	                print HAPFILE " >= ".($NumberOfMissingSNPs+1)."; \n";
	            }
	        }
        }
    }

    # close files

    close HAPFILE;

}  # END OF SUB GENERATE_FILE()



# parsing the results from lp_solve to an array
#

sub parsing {

    my($outputfile) = @_;
    my $i;
    my $j;
    my $x;
    my $a;

    unless (open (OUTFILE, "$outputfile"))
    {
        print "Can not open the file $outputfile !\n";
        exit;
    }

    # filter out the first 4 lines
    
    $a=<OUTFILE>;
    
    # check if this problem is infeasible
    if(($a =~ /infeasible/) || (!$a)){
    #if($a =~ /infeasible/)
    	$x[0] = "infeasible";
    	return @x;
    }
    
    $a=<OUTFILE>;
    $a=<OUTFILE>;
    $a=<OUTFILE>;

    while (<OUTFILE>)
    {
        chop;
        @a = split /\s+/, $_;
        $j = substr($a[0], 2, length($a[0])-2);
#       print "--$a[0]--$a[1]--$j--\n";
        $x[$j-1]=$a[1];
    }
    
    close OUTFILE;
    return @x;
}
